'use strict';

const debug = require('debug')('StyleExt');

const PLUGIN = 'StyleExtHtmlWebpackPlugin';
const EVENT = 'html-webpack-plugin-alter-asset-tags';
const CSS_REGEX = /\.css$/;

const error = msg => {
  const err = new Error(msg);
  err.name = PLUGIN + 'Error';
  throw err;
};

const findGeneratedCssFile = (matchFn, compilation) => {
  const filenames = Object.keys(compilation.assets);
  for (let filename of filenames) {
    if (matchFn(filename)) {
      debug(`${EVENT}: CSS file in compilation: '${filename}'`);
      return filename;
    }
  }
  error(`${PLUGIN} could not find ExtractTextWebpackPlugin\'s generated .css file; available files: '${filenames.join()}'`);
};

const validatePassedCssFilename = (cssFilename, compilation) => {
  const matchFn = (filename) => filename === cssFilename;
  return findGeneratedCssFile(matchFn, compilation);
};

const identifyCssFilename = (compilation) => {
  const matchFn = (filename) => CSS_REGEX.test(filename);
  return findGeneratedCssFile(matchFn, compilation);
};

const generateReplacementStyleTag = (cssFilename, compilation) => {
  const css = compilation.assets[cssFilename].source();
  debug(`${EVENT}: CSS in compilation: ${css}`);
  return {
    tagName: 'style',
    closeTag: true,
    innerHTML: css
  };
};

const handlePublicPath = (cssFilename, compilation) => {
  const prefix = compilation.outputOptions.publicPath;
  if (prefix) {
    cssFilename = prefix + cssFilename;
  }
  return cssFilename;
};

const replaceLinkTag = (stylePath, tags, replacementTag) => {
  for (let index = 0; index < tags.length; index++) {
    if (isCssLinkTag(tags[index], stylePath)) {
      tags[index] = replacementTag;
      debug(`${EVENT} replaced <link> with <style>`);
      return;
    }
  }
  error(`S{PLUGIN} could not find <link> element to replace; available head elements: ${JSON.stringify(tags)}`);
};

const isCssLinkTag = (tagDefinition, stylePath) => {
  if (!tagDefinition) return false;
  if (tagDefinition.tagName !== 'link') return false;
  if (!tagDefinition.attributes) return false;
  if (!tagDefinition.attributes.href) return false;
  if (tagDefinition.attributes.href !== stylePath) return false;
  debug(`${EVENT}: link element found for style path '${stylePath}'`);
  return true;
};

class StyleExtHtmlWebpackPlugin {

  constructor (filename) {
    this.enabled = (filename !== false);
    this.cssFilename = (typeof filename === 'string') ? filename : undefined;
    debug(`constructor: enabled=${this.enabled}, filename=${this.cssFilename}`);
  }

  apply (compiler) {
    if (this.enabled) {
      let cssFilename = this.cssFilename;

      // remove <link> and add <style>
      compiler.plugin('compilation', (compilation) => {
        compilation.plugin(EVENT, (pluginArgs, callback) => {
          try {
            cssFilename = (cssFilename)
              ? validatePassedCssFilename(cssFilename, compilation)
              : identifyCssFilename(compilation);
            const replacementTag = generateReplacementStyleTag(cssFilename, compilation);
            const stylePath = handlePublicPath(cssFilename, compilation);
            replaceLinkTag(stylePath, pluginArgs.head, replacementTag);
            callback(null, pluginArgs);
          } catch (err) {
            callback(err);
          }
        });
      });

      // delete external stylesheet here
      compiler.plugin('emit', (compilation, callback) => {
        if (cssFilename) {
          delete compilation.assets[cssFilename];
          debug(`emit: asset '${this.cssFilename}' deleted`);
          callback();
        }
      });
    }
  }

  /**
   * Guard against pre v3 configurations
   */
  static inline (loaders) {
    error(`${PLUGIN} legacy configuration detected - please see https://github.com/numical/style-ext-html-webpack-plugin for how to configure v3.x+`);
  }
}

module.exports = StyleExtHtmlWebpackPlugin;
